#include <admodel.h>

#include <df1b2fun.h>

#include <adrndeff.h>

  extern "C"  {
    void ad_boundf(int i);
  }
#include <union.htp>

  df1b2_parameters * df1b2_parameters::df1b2_parameters_ptr=0;
  model_parameters * model_parameters::model_parameters_ptr=0;
model_data::model_data(int argc,char * argv[]) : ad_comm(argc,argv)
{
  n.allocate("n");
  y.allocate(1,n,"y");
  race.allocate(1,n,"race");
  sex.allocate(1,n,"sex");
  south.allocate(1,n,"south");
  m1.allocate("m1");
  X_B1.allocate(1,n,1,m1,"X_B1");
  J1.allocate(1,n,1,2,"J1");
  m2.allocate("m2");
  X_B2.allocate(1,n,1,m2,"X_B2");
  J2.allocate(1,n,1,2,"J2");
  m3.allocate("m3");
  X_B3.allocate(1,n,1,m3,"X_B3");
  J3.allocate(1,n,1,2,"J3");
}

model_parameters::model_parameters(int sz,int argc,char * argv[]) : 
 model_data(argc,argv) , function_minimizer(sz)
{
  model_parameters_ptr=this;
  initializationfunction();
  b_race.allocate(1,3,-3,3,1,"b_race");
  b_sex.allocate(-3,3,1,"b_sex");
  b_south.allocate(-3,3,1,"b_south");
  lambda1.allocate(.0001,10,2,"lambda1");
  lambda2.allocate(.0001,40,2,"lambda2");
  lambda3.allocate(.0001,10,2,"lambda3");
  f_wages.allocate(1,n,"f_wages");
  #ifndef NO_AD_INITIALIZE
    f_wages.initialize();
  #endif
  f_age.allocate(1,n,"f_age");
  #ifndef NO_AD_INITIALIZE
    f_age.initialize();
  #endif
  f_ed.allocate(1,n,"f_ed");
  #ifndef NO_AD_INITIALIZE
    f_ed.initialize();
  #endif
  u1.allocate(1,m1,2,"u1");
  u2.allocate(1,m2,2,"u2");
  u3.allocate(1,m3,2,"u3");
  g.allocate("g");  /* ADOBJECTIVEFUNCTION */
}
void model_parameters::userfunction(void)
{
  int i,j;
  g = 0;
  // Add 1rst order difference penalties 
  // Note that the smothing parameter lambda enters below
  for(j=2;j<=m1;j++)
    g -= -0.5*square(u1(j) - u1(j-1));
  for(j=2;j<=m2;j++)
    g -= -0.5*square(u2(j) - u2(j-1));
  for(j=2;j<=m3;j++)
    g -= -0.5*square(u3(j) - u3(j-1));
  // Center the splines at zero to make the model identifiable.
  g -= -0.5*(square(mean(u1)) + square(mean(u2)) + square(mean(u3)));
  f_wages = 0;
  f_age = 0;
  f_ed = 0;
  // Calculate the spline functions from the B-spline bases.
  for(i=1;i<=n;i++)
  {
    for(j=J1(i,1);j<=J1(i,2);j++)
      f_wages(i) +=  X_B1(i,j)*u1(j);
  
    for(j=J2(i,1);j<=J2(i,2);j++)
      f_age(i)   +=  X_B2(i,j)*u2(j);
    for(j=J3(i,1);j<=J3(i,2);j++)
      f_ed(i)    +=  X_B3(i,j)*u3(j);
  }
  // Scale up the splines
  f_wages /= lambda1;
  f_age /= lambda2;
  f_ed /= lambda3;
  // Bernoulli likelhood
  for(i=1;i<=n;i++)
  {
    dvariable eta = b_race(race(i)) + b_sex*sex(i) + b_south*south(i) + f_wages(i) + f_age(i) + f_ed(i); 
    g -= y(i)*eta - log(1+mfexp(eta));      
  }
  
}

void model_parameters::report()
{
 adstring ad_tmp=initial_params::get_reportfile_name();
  ofstream report((char*)(adprogram_name + ad_tmp));
  if (!report)
  {
    cerr << "error trying to open report file"  << adprogram_name << ".rep";
    return;
  }
  report << f_wages << endl;
  report << f_age << endl;
  report << f_ed << "\n\n" << endl;
}

void model_parameters::preliminary_calculations(void)
{

  admaster_slave_variable_interface(*this);
  cout << setprecision(4);
}
  long int arrmblsize=0;

int main(int argc,char * argv[])
{
  ad_set_new_handler();
  ad_exit=&ad_boundf;
  arrmblsize = 4000000L;
  gradient_structure::set_GRADSTACK_BUFFER_SIZE(3000000);
  gradient_structure::set_CMPDIF_BUFFER_SIZE(200000);
  gradient_structure::set_MAX_NVAR_OFFSET(130000);
    gradient_structure::set_NO_DERIVATIVES();
    gradient_structure::set_YES_SAVE_VARIABLES_VALUES();
  #if defined(__GNUDOS__) || defined(DOS386) || defined(__DPMI32__)  || \
     defined(__MSVC32__)
      if (!arrmblsize) arrmblsize=150000;
  #else
      if (!arrmblsize) arrmblsize=25000;
  #endif
    df1b2variable::noallocate=1;
    df1b2_parameters mp(arrmblsize,argc,argv);
    mp.iprint=10;

    function_minimizer::random_effects_flag=1;
    df1b2variable::noallocate=0;
    mp.preliminary_calculations();
    mp.computations(argc,argv);
    return 0;
}

extern "C"  {
  void ad_boundf(int i)
  {
    // so we can stop here
    exit(i);
  }
}

model_data::~model_data()
{}

model_parameters::~model_parameters()
{}

void model_parameters::final_calcs(void){}

void model_parameters::set_runtime(void){}

#ifdef _BORLANDC_
  extern unsigned _stklen=10000U;
#endif


#ifdef __ZTC__
  extern unsigned int _stack=10000U;
#endif

void df1b2_parameters::user_function(void)
{
  int i,j;
  g = 0;
  // Add 1rst order difference penalties 
  // Note that the smothing parameter lambda enters below
  for(j=2;j<=m1;j++)
    g -= -0.5*square(u1(j) - u1(j-1));
  for(j=2;j<=m2;j++)
    g -= -0.5*square(u2(j) - u2(j-1));
  for(j=2;j<=m3;j++)
    g -= -0.5*square(u3(j) - u3(j-1));
  // Center the splines at zero to make the model identifiable.
  g -= -0.5*(square(mean(u1)) + square(mean(u2)) + square(mean(u3)));
  f_wages = 0;
  f_age = 0;
  f_ed = 0;
  // Calculate the spline functions from the B-spline bases.
  for(i=1;i<=n;i++)
  {
    for(j=J1(i,1);j<=J1(i,2);j++)
      f_wages(i) +=  X_B1(i,j)*u1(j);
  
    for(j=J2(i,1);j<=J2(i,2);j++)
      f_age(i)   +=  X_B2(i,j)*u2(j);
    for(j=J3(i,1);j<=J3(i,2);j++)
      f_ed(i)    +=  X_B3(i,j)*u3(j);
  }
  // Scale up the splines
  f_wages /= lambda1;
  f_age /= lambda2;
  f_ed /= lambda3;
  // Bernoulli likelhood
  for(i=1;i<=n;i++)
  {
    df1b2variable eta = b_race(race(i)) + b_sex*sex(i) + b_south*south(i) + f_wages(i) + f_age(i) + f_ed(i); 
    g -= y(i)*eta - log(1+mfexp(eta));      
  }
  
}

void df1b2_pre_parameters::setup_quadprior_calcs(void) 
{ 
df1b2_gradlist::set_no_derivatives(); 
quadratic_prior::in_qp_calculations=1; 
}  

void df1b2_pre_parameters::begin_df1b2_funnel(void) 
{ 
(*re_objective_function_value::pobjfun)=0; 
other_separable_stuff_begin(); 
f1b2gradlist->reset();  
if (!quadratic_prior::in_qp_calculations) 
{ 
df1b2_gradlist::set_yes_derivatives();  
} 
funnel_init_var::allocate_all();  
}  

void df1b2_pre_parameters::end_df1b2_funnel(void) 
{  
lapprox->do_separable_stuff(); 
other_separable_stuff_end(); 
} 

void model_parameters::begin_df1b2_funnel(void) 
{ 
if (lapprox)  
{  
{  
begin_funnel_stuff();  
}  
}  
}  

void model_parameters::end_df1b2_funnel(void) 
{  
if (lapprox)  
{  
end_df1b2_funnel_stuff();  
}  
} 
void df1b2_parameters::allocate(void) 
{
  b_race.allocate(1,3,-3,3,1,"b_race");
  b_sex.allocate(-3,3,1,"b_sex");
  b_south.allocate(-3,3,1,"b_south");
  lambda1.allocate(.0001,10,2,"lambda1");
  lambda2.allocate(.0001,40,2,"lambda2");
  lambda3.allocate(.0001,10,2,"lambda3");
  f_wages.allocate(1,n,"f_wages");
  #ifndef NO_AD_INITIALIZE
    f_wages.initialize();
  #endif
  f_age.allocate(1,n,"f_age");
  #ifndef NO_AD_INITIALIZE
    f_age.initialize();
  #endif
  f_ed.allocate(1,n,"f_ed");
  #ifndef NO_AD_INITIALIZE
    f_ed.initialize();
  #endif
  u1.allocate(1,m1,2,"u1");
  u2.allocate(1,m2,2,"u2");
  u3.allocate(1,m3,2,"u3");
  g.allocate("g");  /* ADOBJECTIVEFUNCTION */
}
